#!/usr/bin/env python3
#apt-get install gir1.2-gstreamer-1.0 gstreamer-1.0-plugins-{base,good}
import argparse
import collections
import logging
import math
import re
import sys
import subprocess
try:
    import gi
    gi.require_version('Gst','1.0')
    from gi.repository import GObject
    from gi.repository import Gst
except ImportError:
    print("Can't import module: %s. it may not be available for this"
          "version of Python, which is: " % sys.exc_info()[1], file=sys.stderr)
    print((sys.version), file=sys.stderr)
    sys.exit(127)


#Frequency bands for FFT
BINS = 256
#How often to take a sample and do FFT on it.
FFT_INTERVAL = 100000000  # In nanoseconds, so this is every 1/10th second
#Sampling frequency. The effective maximum frequency we can analyze is
#half of this (see Nyquist's theorem)
SAMPLING_FREQUENCY = 44100
#The default test frequency should be in the middle of the frequency band
#that delimits the first and second thirds of the frequency range.
#That gives a not-so-ear-piercing tone and should ensure there's no
#spillout to neighboring frequency bands.
DEFAULT_TEST_FREQUENCY = (SAMPLING_FREQUENCY / (2 * BINS)) * int(BINS / 3) - \
                         (SAMPLING_FREQUENCY / (2 * BINS)) / 2
#only sample a signal when peak level is in this range (in dB attenuation,
#0 means no attenuation (and horrible clipping).
REC_LEVEL_RANGE = (-2.0, -12.0)
#For our test signal to be considered present, it has to be this much higher
#than the average of the rest of the frequencies (to ensure we have a nice,
#clear peak). This is in dB.
MAGNITUDE_THRESHOLD = -5.0


class PIDController(object):
    def __init__(self, Kp, Ki, Kd, setpoint=0):
        self.setpoint = setpoint
        self.Kp = Kp
        self.Ki = Ki
        self.Kd = Kd
        self._integral = 0
        self._previous_error = 0
        self._change_limit = 0

    def input_change(self, process_feedback, dt):
        """ Calculates desired input value change.

            Based on process feedback and time inteval (dt).
        """
        error = self.setpoint - process_feedback
        self._integral = self._integral + (error * dt)
        derivative = (error - self._previous_error) / dt
        self._previous_error = error
        input_change = (self.Kp * error) + \
                       (self.Ki * self._integral) + \
                       (self.Kd * derivative)
        if self._change_limit and abs(input_change) > abs(self._change_limit):
            sign = input_change / abs(input_change)
            input_change = sign * self._change_limit
        return input_change

    def set_change_limit(self, limit):
        """Ensures that input value changes are lower than limit.

           Setting limit of zero disables this.
        """
        self._change_limit = limit


class PAVolumeController(object):
    pa_types = {'input': 'source', 'output': 'sink'}

    def __init__(self, type, method=None, logger=None):
        """Initializes the volume controller.

           Arguments:
           type: either input or output
           method: a method that will run a command and return pulseaudio
           information in the described format, as a single string with
           line breaks (to be processed with str.splitlines())

        """
        self.type = type
        self._volume = None
        self.identifier = None
        self.method = method
        if not isinstance(method, collections.Callable):
            self.method = self._pactl_output
        self.logger = logger

    def set(self, volume):
        if not 0 <= volume <= 100:
            return False
        if not self.identifier:
            return False
        command = ['pactl',
                   'set-%s-volume' % (self.pa_types[self.type]),
                   str(self.identifier[0]),
                   str(int(volume)) + "%"]
        if False == self.method(command):
            return False
        self._volume = volume
        return True

    def get(self):
        if not self.identifier:
            return None
        return self._volume

    def mute(self, mute):
        mute = str(int(mute))
        if not self.identifier:
            return False
        command = ['pactl',
                   'set-%s-mute' % (self.pa_types[self.type]),
                   str(self.identifier[0]),
                   mute]
        if False == self.method(command):
            return False
        return True

        

    def get_identifier(self):
        if self.type:
            self.identifier = self._get_identifier_for(self.type)
            if self.identifier and self.logger:
                message = "Using PulseAudio identifier %s (%s) for %s" %\
                       (self.identifier + (self.type,))
                self.logger.info(message)
            return self.identifier

    def _get_identifier_for(self, type):
        """Gets default PulseAudio identifier for given type.

           Arguments:
           type: either input or output

           Returns:
           A tuple: (pa_id, pa_description)

        """

        if type not in self.pa_types:
            return None
        command = ['pactl', 'list', self.pa_types[type] + "s", 'short']

        #Expect lines of this form (field separator is tab):
        #<ID>\t<NAME>\t<MODULE>\t<SAMPLE_SPEC_WITH_SPACES>\t<STATE>
        #What we need to return is the ID for the first element on this list
        #that does not contain auto_null or monitor.
        pa_info = self.method(command)
        valid_elements = None

        if pa_info:
            valid_elements = [(int(i.split()[0]), i.split()[1]) \
                        for i in pa_info.splitlines() \
                        if not re.match('.*monitor.*', i) \
                        and not re.match('.*auto_null.*', i)]
        if not valid_elements:
            if self.logger:
                self.logger.error("No valid PulseAudio elements"
                                  " for %s" % (self.type))
            return None
        return valid_elements[0]

    def _pactl_output(self, command):
        try:
            return subprocess.check_output(command, universal_newlines=True)
        except (subprocess.CalledProcessError, AttributeError):
            return False


class FileDumper(object):
    def write_to_file(self, filename, data):
        try:
            with open(filename, "wb") as f:
                for i in data:
                    print(i, file=f)
            return_value = True
        except (TypeError, IOError):
            return_value = False
        return return_value


class SpectrumAnalyzer(object):
    def __init__(self, points, sampling_frequency=44100,
                 wanted_samples=20):
        self.spectrum = [0] * points
        self.number_of_samples = 0
        self.wanted_samples = wanted_samples
        self.sampling_frequency = sampling_frequency
        #Frequencies should contain *real* frequency which is half of
        #the sampling frequency
        self.frequencies = [((sampling_frequency / 2.0) / points) * i
                            for i in range(points)]

    def _average(self):
        return sum(self.spectrum) / len(self.spectrum)

    def sample(self, sample):
        if len(sample) != len(self.spectrum):
            return
        self.spectrum = [((old * self.number_of_samples) + new) /
                         (self.number_of_samples + 1)
                         for old, new in zip(self.spectrum, sample)]
        self.number_of_samples += 1

    def frequencies_over_average(self, threshold=0.0):
        return [i for i in range(len(self.spectrum)) \
                if self.spectrum[i] >= self._average() - threshold]

    def frequency_band_for(self, frequency):
        """Convenience function to tell me which band
           a frequency is contained in
        """
        #Note that actual frequencies are half of what the sampling
        #frequency would tell us. If SF is 44100 then maximum actual
        #frequency is 22050, and if I have 10 frequency bins each will
        #contain only 2205 Hz, not 4410 Hz.
        max_frequency = self.sampling_frequency / 2
        if frequency > max_frequency or frequency < 0:
            return None
        band = float(frequency) / (max_frequency / len(self.spectrum))
        return int(math.ceil(band)) - 1

    def frequencies_for_band(self, band):
        """Convenience function to tell me the delimiting frequencies
           for a band
        """
        if band >= len(self.spectrum) or band < 0:
            return None
        lower = self.frequencies[band]
        upper = lower + ((self.sampling_frequency / 2.0) / len(self.spectrum))
        return (lower, upper)

    def sampling_complete(self):
        return self.number_of_samples >= self.wanted_samples


class GStreamerRawAudioRecorder(object):
    def __init__(self):
        self.raw_buffers = []

    def buffer_handler(self, sink):
        buffer = sink.emit('pull-sample')
        self.raw_buffers.append(buffer.get_buffer())

    def get_raw_audio(self):
        return ''.join(self.raw_buffers)


class GStreamerMessageHandler(object):
    def __init__(self, rec_level_range, logger, volumecontroller,
                 pidcontroller, spectrum_analyzer):
        """Initializes the message handler. It knows how to handle
           spectrum and level gstreamer messages.

           Arguments:
           rec_level_range: tuple with acceptable recording level
                            ranges
           logger: logging object with debug, info, error methods.
           volumecontroller: an instance of VolumeController to use
                             to adjust RECORDING level
           pidcontroller: a PID controller instance which helps control
                          volume
           spectrum_analyzer: instance of SpectrumAnalyzer to collect
                              data from spectrum messages

        """
        self.current_level = sys.maxsize
        self.logger = logger
        self.pid_controller = pidcontroller
        self.rec_level_range = rec_level_range
        self.spectrum_analyzer = spectrum_analyzer
        self.volume_controller = volumecontroller

    def set_quit_method(self, method):
        """ Method that will be called when sampling is complete."""
        self._quit_method = method

    def bus_message_handler(self, bus, message):
        if message.type == Gst.MessageType.ELEMENT:
            message_name = message.get_structure().get_name()
            if message_name == 'spectrum':
                print( dir(message.get_structure().get_value('magnitude')))
                import pdb; pdb.set_trace()
                sys.exit(1)
                fft_magnitudes = message.get_structure()['magnitude']
                self.spectrum_method(self.spectrum_analyzer, fft_magnitudes)

            if message_name == 'level':
                #peak_value is our process feedback
                peak_value = message.get_structure()['peak'][0]
                self.level_method(peak_value, self.pid_controller,
                                  self.volume_controller)

    #Adjust recording level
    def level_method(self, level, pid_controller, volume_controller):
        #If volume controller doesn't return a valid volume,
        #we can't control it :(
        current_volume = volume_controller.get()
        if current_volume == None:
            self.logger.error("Unable to control recording volume."
                              "Test results may be wrong")
            return
        self.current_level = level
        change = pid_controller.input_change(level, 0.10)
        if self.logger:
            self.logger.debug("Peak level: %(peak_level).2f, "
                         "volume: %(volume)d%%, Volume change: %(change)f%%" %
                      {'peak_level': level,
                       'change': change,
                       'volume': current_volume})
        volume_controller.set(current_volume + change)

    #Only sample if level is within the threshold
    def spectrum_method(self, analyzer, spectrum):
        if self.rec_level_range[1] <= self.current_level \
           or self.current_level <= self.rec_level_range[0]:
            self.logger.debug("Sampling, recorded %d samples" %
                               analyzer.number_of_samples)
            analyzer.sample(spectrum)
        if analyzer.sampling_complete() and self._quit_method:
            self.logger.info("Sampling complete, ending process")
            self._quit_method()


class GstAudioObject(object):
    def __init__(self):
        self.class_name = self.__class__.__name__

    def _set_state(self, state, description):
        self.pipeline.set_state(state)
        message = "%s: %s" % (self.class_name, description)
        if self.logger:
            self.logger.info(message)

    def start(self):
        self._set_state(Gst.State.PLAYING, "Starting")

    def stop(self):
        self._set_state(Gst.State.NULL, "Stopping")


class Player(GstAudioObject):
    def __init__(self, frequency=DEFAULT_TEST_FREQUENCY, logger=None):
        super(Player, self).__init__()
        self.pipeline_description = ("audiotestsrc wave=sine freq=%s "
                                "! audioconvert "
                                "! audioresample "
                                "! autoaudiosink" % int(frequency))
        self.logger = logger
        if self.logger:
            self.logger.debug(self.pipeline_description)
        self.pipeline = Gst.parse_launch(self.pipeline_description)


class Recorder(GstAudioObject):
    def __init__(self, bins=BINS, sampling_frequency=SAMPLING_FREQUENCY,
                 fft_interval=FFT_INTERVAL, logger=None):
        super(Recorder, self).__init__()
        pipeline_description = ('''autoaudiosrc
        ! queue
        ! level message=true
        ! audioconvert
        ! audio/x-raw,channels=1,rate=(int)%(rate)s
        ! audioresample
        ! spectrum interval=%(fft_interval)s bands = %(bands)s
        ! wavenc
        ! appsink name=recordersink emit-signals=true''' %
        {'bands': bins,
         'rate': sampling_frequency,
         'fft_interval': fft_interval})
        self.logger = logger
        if self.logger:
            self.logger.debug(pipeline_description)
        self.pipeline = Gst.parse_launch(pipeline_description)

    def register_message_handler(self, handler_method):
        if self.logger:
            message = "Registering message handler: %s" % handler_method
            self.logger.debug(message)
        self.bus = self.pipeline.get_bus()
        self.bus.add_signal_watch()
        self.bus.connect('message', handler_method)

    def register_buffer_handler(self, handler_method):
        if self.logger:
            message = "Registering buffer handler: %s" % handler_method
            self.logger.debug(message)
        self.sink = self.pipeline.get_by_name('recordersink')
        self.sink.connect('new-sample', handler_method)


def process_arguments():
    description = """
        Plays a single frequency through the default output, then records on 
        the default input device. Analyzes the recorded signal to test for  
        presence of the played frequency, if present it exits with success.
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("-t", "--time",
            dest='test_duration',
            action='store',
            default=30,
            type=int,
            help="""Maximum test duration, default %(default)s seconds.
                    It may exit sooner if it determines it has enough data.""")
    parser.add_argument("-a", "--audio",
            action='store',
            default=None,
            type=str,
            help="File to save recorded audio in .wav format")
    parser.add_argument("-v", "--verbose",
            action='store_true',
            default=False,
            help="Be verbose")
    parser.add_argument("-d", "--debug",
            action='store_true',
            default=False,
            help="Debugging output")
    parser.add_argument("-f", "--frequency",
            action='store',
            default=DEFAULT_TEST_FREQUENCY,
            type=int,
            help="Frequency for test signal, default %(default)s Hz")
    parser.add_argument("-u", "--spectrum",
            action='store',
            type=str,
            help="""File to save spectrum information for plotting
                    (one frequency/magnitude pair per line)""")
    return parser.parse_args()


#
def main():
    #Get arguments.
    args = process_arguments()

    #Setup logging
    level = logging.ERROR
    if args.debug:
        level = logging.DEBUG
    if args.verbose:
        level = logging.INFO
    logging.basicConfig(level=level)
    try:
        Gst.init(None)
        #Launches recording pipeline. I need to hook up into the gst
        #messages.
        recorder = Recorder(logger=logging)
        #Just launches the playing pipeline
        player = Player(frequency=args.frequency, logger=logging)
    except GObject.GError as e:
        logging.critical("Unable to initialize GStreamer pipelines")
        print(sys.exc_info()[1])
        sys.exit(127)

    #This just receives a process feedback and tells me how much to change to
    #achieve the setpoint
    pidctrl = PIDController(Kp=0.7, Ki=.01, Kd=0.01,
                            setpoint=REC_LEVEL_RANGE[0])
    pidctrl.set_change_limit(5)
    #This  gathers spectrum data. 
    analyzer = SpectrumAnalyzer(points=BINS,
                                sampling_frequency=SAMPLING_FREQUENCY)
    #this receives 'buffer' messages and gathers raw audio data
    rawaudio = GStreamerRawAudioRecorder()

    #Volume controllers actually set volumes for their device types.
    #we should at least issue a warning
    recorder.volumecontroller = PAVolumeController(type='input',
                                                   logger=logging)
    if not recorder.volumecontroller.get_identifier():
        logging.warning("Unable to get input volume control identifier. "
                       "Test results will probably be invalid")
    recorder.volumecontroller.set(0)

    player.volumecontroller = PAVolumeController(type='output',
                                                 logger=logging)
    if not player.volumecontroller.get_identifier():
        logging.warning("Unable to get output volume control identifier. "
                       "Test results will probably be invalid")
    player.volumecontroller.set(70)
    player.volumecontroller.mute(False)

    #This handles the messages from gstreamer and orchestrates
    #the passed volume controllers, pid controller and spectrum analyzer
    #accordingly.
    gmh = GStreamerMessageHandler(rec_level_range=REC_LEVEL_RANGE,
                                  logger=logging,
                                  volumecontroller=recorder.volumecontroller,
                                  pidcontroller=pidctrl,
                                  spectrum_analyzer=analyzer)

    #I need to tell the recorder which method will handle messages.
    recorder.register_message_handler(gmh.bus_message_handler)
    recorder.register_buffer_handler(rawaudio.buffer_handler)

    #Create the loop and add a few triggers 
    GObject.threads_init()
    loop = GObject.MainLoop()
    GObject.timeout_add_seconds(0, player.start)
    GObject.timeout_add_seconds(0, recorder.start)
    GObject.timeout_add_seconds(args.test_duration, loop.quit)

    # Tell the gmh which method to call when enough samples are collected
    gmh.set_quit_method(loop.quit)

    loop.run()

    #When the loop ends, set things back to reasonable states
    player.stop()
    recorder.stop()
    player.volumecontroller.set(50)
    recorder.volumecontroller.set(10)

    #See if data gathering was successful.
    test_band = analyzer.frequency_band_for(args.frequency)
    candidate_bands = analyzer.frequencies_over_average(MAGNITUDE_THRESHOLD)
    if test_band in candidate_bands:
        freqs_for_band = analyzer.frequencies_for_band(test_band)
        print("PASS: Test frequency of %s in band (%.2f, %.2f) "
              "which had a magnitude higher than the average" %
            ((args.frequency,) + freqs_for_band))
        return_value = 0
    else:
        print("FAIL: Test frequency of %s is not in one of the "
              "bands with higher-than-average magnitude" % args.frequency)
        return_value = 1
    #Is the microphone broken?
    if len(set(analyzer.spectrum)) <= 1:
        print("WARNING: Microphone seems broken, didn't even record ambient"
              " noise")

    #Write some files to disk for later analysis
    if args.audio:
        if args.verbose:
            print("Saving recorded audio as %s" % args.audio)
        if not FileDumper().write_to_file(args.audio, [rawaudio.get_raw_audio()]):
            print("Couldn't save recorded audio", file=sys.stderr)

    if args.spectrum:
        if args.verbose:
            print("Saving spectrum data for plotting as %s" % args.spectrum)
        if not FileDumper().write_to_file(args.spectrum,
                                       ["%s,%s" % t for t in 
                                        zip(analyzer.frequencies, 
                                            analyzer.spectrum)]):
            print("Couldn't save spectrum data for plotting", file=sys.stderr)

    return return_value

if __name__ == "__main__":
    sys.exit(main())
